from yostlabs.tss3.api import ThreespaceSensor
from xml.dom import minidom, Node

from typing import Callable

class ThreespaceFirmwareUploader:

    def __init__(self, sensor: ThreespaceSensor, file_path: str = None, percentage_callback: Callable[[int],None] = None, verbose: bool = False):
        self.sensor = sensor
        self.set_firmware_path(file_path)
        self.verbose = verbose

        self.percent_complete = 0
        self.callback = percentage_callback

    def set_firmware_path(self, file_path: str):
        if file_path is None: 
            self.firmware = None
            return
        self.firmware = minidom.parse(file_path)

    def set_percent_callback(self, callback: Callable[[float],None]):
        self.callback = callback

    def set_verbose(self, verbose: bool):
        self.verbose = verbose

    def get_percent_done(self):
        return self.percent_complete
    
    def __set_percent_complete(self, percent: float):
        self.percent_complete = percent
        if self.callback:
            self.callback(percent)

    def log(self, *args):
        if not self.verbose: return
        print(*args)

    def upload_firmware(self):
        self.percent_complete = 0
        if not self.sensor.in_bootloader:
            self.sensor.enterBootloader()
        self.__set_percent_complete(5)
        
        boot_info = self.sensor.bootloader_get_info()

        root = self.firmware.firstChild
        for c in root.childNodes:
            if c.nodeType == Node.ELEMENT_NODE:
                name = c.nodeName
                if name == "SetAddr":
                    self.log("Write S")
                    error = self.sensor.bootloader_erase_firmware()
                    if error:
                        self.log("Failed to erase firmware:", error)
                    else:
                        self.log("Successfully erased firmware")
                    self.__set_percent_complete(20)
                elif name == "MemProgC":
                    mem = bytes.fromhex(c.firstChild.nodeValue)
                    self.log("Attempting to program", len(mem), "bytes to the chip")
                    cpos = 0
                    while cpos < len(mem):
                        memchunk = mem[cpos : min(len(mem), cpos + boot_info.pagesize)]
                        error = self.sensor.bootloader_prog_mem(memchunk)
                        if error:
                            self.log("Failed upload:", error)
                        else:
                            self.log("Wrote", len(memchunk), "bytes successfully to offset", cpos)
                        cpos += len(memchunk)
                        self.__set_percent_complete(20 + cpos / len(mem) * 79)
                elif name == "Run":
                    self.log("Resetting with new firmware.")
                    self.sensor.bootloader_boot_firmware()
                    self.__set_percent_complete(100)


from yostlabs.communication.base import *
import socket
import time
from typing import Callable

class ThreespaceSocketComClass(ThreespaceComClass):

    """
    Inheriting classes must implement auto_detect() functionality.
    Should also implement 'name' functionality
    """

    def __init__(self, sock: socket.socket, *connection_params, connection_timeout=None):
        self.socket = sock
        self._opened = False

        self.buffer = bytearray()
        self.__timeout = 2
        self.__connection_timeout = connection_timeout
        self.__connection_params = connection_params

    @property
    def timeout(self) -> float:
        return self.__timeout
    
    @timeout.setter
    def timeout(self, timeout: float):
        self.__timeout = timeout

    def open(self):
        if self._opened: return True
        try:
            if self.__connection_timeout is None:
                self.socket.setblocking(True)
            else:
                self.socket.settimeout(self.__connection_timeout)
            self.socket.connect(*self.__connection_params)
        except Exception as e:
            try:
                self.socket.close()
            except: pass
            print(e)
            return False
        self._opened = True
        return True

    def close(self):
        if not self._opened: return
        self.socket.close()
        self._opened = False

    def check_open(self):
        return self._opened

    def write(self, bytes: bytes):
        self.socket.send(bytes)

    def read(self, num_bytes: int):
        self.__update_while(lambda: len(self.buffer) < num_bytes)
        amount = min(len(self.buffer), num_bytes)
        result = self.buffer[:amount]
        del self.buffer[:amount]
        return result
    
    def peek(self, num_bytes: int):
        self.__update_while(lambda: len(self.buffer) < num_bytes)
        amount = min(len(self.buffer), num_bytes)
        return self.buffer[:amount]
        
    def read_until(self, expected: bytes):
        self.__update_while(lambda: expected not in self.buffer)
        if expected in self.buffer:
            length = self.buffer.index(expected) + len(expected)
            result = self.buffer[:length]
            del self.buffer[:length]
        else:
            result = self.buffer.copy()
            self.buffer.clear()
        return result
    
    def peek_until(self, expected: bytes, max_length: int = None):
        self.__update_while(lambda: expected not in self.buffer and (max_length is None or len(self.buffer) < max_length))
        if expected in self.buffer:
            length = self.buffer.index(expected) + len(expected)
            if max_length is not None:
                length = min(length, max_length)
            result = self.buffer[:length]
        else:
            length = len(self.buffer)
            if max_length is not None:
                length = min(length, max_length)
            result = self.buffer[length]
        return result

    def __update_while(self, condition: Callable):
        start_time = time.perf_counter()
        while condition():
            self.__update_buffer()
            if time.perf_counter() - start_time >= self.timeout:
                return

    def __update_buffer(self, max_bytes: int = 1000):
        self.socket.setblocking(False)
        try:
            self.buffer += self.socket.recv(max_bytes)
        except BlockingIOError:
            return False
        self.socket.settimeout(self.timeout)
        return True

    @property
    def length(self):
        while self.__update_buffer(): pass
        return len(self.buffer)

    @property
    def reenumerates(self) -> bool:
        return False


from yostlabs.tss3.api import *
import math

class ThreespaceBufferInputStream(ThreespaceInputStream):
    """
    Default Input Stream for the binary parser.
    Ignores timeout since this is only used synchronously for now.
    """

    def __init__(self):
        self.buffer = bytearray()

    """Reads specified number of bytes."""
    def read(self, num_bytes) -> bytes:
        num_bytes = min(len(self.buffer), num_bytes)
        result = self.buffer[:num_bytes]
        del self.buffer[:num_bytes]
        return result
    
    def read_all(self):
        return self.read(self.length)

    def read_until(self, expected: bytes) -> bytes:
        if expected not in self.buffer:
            return self.read_all()
        
        length = self.buffer.index(expected) + len(expected)
        result = self.buffer[:length]
        del self.buffer[:length]
        return result

    """Allows reading without removing the data from the buffer"""
    def peek(self, num_bytes) -> bytes:
        num_bytes = min(len(self.buffer), num_bytes)
        return self.buffer[:num_bytes]
    
    def peek_until(self, expected: bytes, max_length=None) -> bytes:
        if expected in self.buffer: #Read until the expected
            length = self.buffer.index(expected) + len(expected)
            if max_length is not None and length > max_length:
                length = max_length
            return self.buffer[:length]
        
        #There is no expected, so read as far as possible
        length = len(self.buffer)
        if max_length is not None:
            length = min(length, max_length)
        return self.buffer[:length]
    
    def readline(self) -> bytes:
        return self.read_until(b"\n")
    
    def peekline(self, max_length=None) -> bytes:
        return self.peek_until(b"\n", max_length=max_length)    
    
    def insert(self, data: bytes):
        self.buffer.extend(data)

    @property
    def length(self) -> int:
        return len(self.buffer)
    
    @property
    def timeout(self) -> float:
        raise NotImplementedError()

    @timeout.setter
    def timeout(self, timeout: float):
        raise NotImplementedError()    

class ThreespaceBinaryParser:
    """
    A class that can be used to parse a stream of binary data
    that could contain multiple different command responses and validates
    the responses to handle misalignment/data corruption.

    Requires all expected responses to have the same header enabled.

    The header should contain the cmd_echo, checksum, and data_length fields
    for full functionality. The lack of any of those fields could limit functionality
    of the parser.

    If cmd_echo is missing, only one command can be registered with the binary parser as it has no way
    of knowing what of verifying what the current incoming response is.

    If checksum is missing, data integrity can not be checked.

    If data_length is missing, commands that do not return a static length may cause blocking operations
    or errors while parsing. (These are planned to be fixed in a future version)

    NOTE: For speed, a custom implementation will be better then this parser class. This parser handles allowing
    multiple commands as well as data validation and misalignment correction. It also formats the data into the ThreespaceCmdResult
    response type. The overhead added because of all these additional checks/calculations can add a significant amount of time
    to processing binary data compared to just reading a known amount of data and instantly unpacking it to a tuple in the desired format.
    """

    COMMAND_EXCEPTIONS = [84, 177] #getStreamingBatch and fileReadBytes need additional info and so need registered via the 

    def __init__(self, data_stream: ThreespaceInputStream = None, verbose=False):
        """
        Parameters
        ----------
        data_stream - (optional) The data stream to use with the Binary Parser. If not supplied, will default to a new ThreespaceBufferInputStream
        """
        self.data_stream = data_stream
        self.registered_commands: dict[int,ThreespaceCommand] = {}

        if self.data_stream is None:
            self.data_stream = ThreespaceBufferInputStream()
        self.header_info = None

        self.__parsing_header: ThreespaceHeader = None  #Used to optimize preventing reading to much by cacheing the header seperately from the cmd data
        self.__parsing_command: ThreespaceCommand = None
        self.__parsing_msg_length: int = None           #Used seperately from the __parsing_header so can handle msg lengths that are static without modifying the header

        self.misaligned = False
        self.verbose = verbose

    def register_command(self, cmd: int|ThreespaceCommand, **kwargs):
        """
        Registers the given cmd number/cmd with the binary parser.

        Some commands may require additional information:
        stream_slots - list[int] Required when registering command 84 (getStreamingBatch) a list of command numbers that are being streamed.
        read_size - 'auto' or int Required when registering a command that requires a given length such as fileReadBytes. If 'auto' will use the header length to determine length.
        """
        if isinstance(cmd, int):
            cmd = threespaceCommandGet(cmd)
            if cmd is None:
                raise ValueError(f"Invalid Cmd {cmd}")
            
        if cmd.info.num in self.registered_commands:
            return False
        
        #These command types are special and need additional info
        if cmd.info.num == THREESPACE_FILE_READ_BYTES_COMMAND_NUM:
            if "read_size" not in kwargs:
                raise ValueError("Missing arguement 'read_size' when registering the fileReadBytes command with the binary parser")
            raise NotImplementedError("The fileReadBytes command has yet to be implemented for the ThreespaceBinaryParser")
        elif cmd.info.num == THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM and not isinstance(cmd, ThreespaceGetStreamingBatchCommand):
            if "stream_slots" not in kwargs:
                raise ValueError("Missing arguement 'stream_slots' when registering the getStreamingBatch command with the binary parser")
            cmd = ThreespaceGetStreamingBatchCommand(kwargs['stream_slots'])

        self.registered_commands[cmd.info.num] = cmd
        return True

    def unregister_command(self, cmd: int|ThreespaceCommand):
        if cmd.info.num not in self.registered_commands:
            return False
        del self.registered_commands[cmd.info.num]
        return True

    def set_header(self, header_info: ThreespaceHeaderInfo):
        self.header_info = header_info

    def insert_data(self, data: bytes):
        """
        Add the given data to the default ThreespaceBufferInputStream.
        This method will raise an exception if used on a different type of InputStream
        """
        if not isinstance(self.data_stream, ThreespaceBufferInputStream):
            raise Exception("Insert data with Binary Parser only valid when using the default data_stream")
        self.data_stream.insert(data)
    
    def parse_message(self) -> ThreespaceCmdResult:
        if self.__parsing_header is None:
            self.__parse_header()
        
        if self.__parsing_command is None:
            return None
        
        return self.__parse_command()

    def __parse_header(self):
        if self.data_stream.length < self.header_info.size:
            return
        
        header = self.data_stream.peek(self.header_info.size)
        header = ThreespaceHeader.from_bytes(header, self.header_info)

        cmd_found = False
        if self.header_info.echo_enabled: #Search for the command to parse
            for command in self.registered_commands.values():
                if header.echo != command.info.num: continue

                #Command matches! Attempt to parse
                self.__parsing_command = command
                self.__parsing_header = header

                cmd_found = True
        else: #Can only parse one command
            if len(self.registered_commands) > 1:
                raise Exception("Only one command type can be parsed when the 'cmd echo' is not enabled in the header")
            self.__parsing_command = list(self.registered_commands.values())[0]
            self.__parsing_header = header
            cmd_found = True

        if cmd_found:
            if self.header_info.length_enabled:
                self.__parsing_msg_length = self.__parsing_header.length
            else:
                self.__parsing_msg_length = self.__parsing_command.info.out_size
            return

        #This header is not related to any command, so it needs skipped
        if self.verbose and not self.misaligned:
            print("Unexpected header:", header)
        self.misaligned = True
        self.data_stream.read(1)

    def __peek_checksum(self):
        header_len = len(self.__parsing_header.raw_binary)
        data = self.data_stream.peek(header_len + self.__parsing_msg_length)[header_len:]
        checksum = sum(data) % 256
        return checksum == self.__parsing_header.checksum

    def __parse_command(self):
        #Not enough data to parse yet
        if self.data_stream.length < self.header_info.size + self.__parsing_msg_length:
            return None
        
        if self.header_info.checksum_enabled and not math.isnan(self.__parsing_msg_length): #Can validate checksum before parsing
            if not self.__peek_checksum():
                #Data corruption/Misalignment error
                if self.verbose and not self.misaligned:
                    print("Checksum mismatch for command", self.__parsing_command.info.num)
                self.misaligned = True
                self.data_stream.read(1)
                self.__parsing_command = None
                self.__parsing_header = None
                return None
        
        #Header and pre validation checksum checks out! Now just parse the actual command result and return it
        header = self.__parsing_header
        self.data_stream.read(len(header.raw_binary)) #Skip these bytes since they are already parsed
        result, raw = self.__parsing_command.read_command(self.data_stream)

        #Validate checksum if couldn't pre-validate due to unknown message length
        if math.isnan(self.__parsing_msg_length) and self.header_info.checksum_enabled:
            checksum = sum(raw) % 256
            if checksum != header.checksum:
                if self.verbose and not self.misaligned:
                    print("Checksum mismatch for command", self.__parsing_command.info.num)
                self.misaligned = True
                self.__parsing_command = None
                self.__parsing_header = None
                return None

        #Reset and return
        self.__parsing_header = None
        self.__parsing_command = None
        self.misaligned = False
        return ThreespaceCmdResult(result, header, data_raw_binary=raw)
        


from yostlabs.tss3.api import ThreespaceSensor
from yostlabs.communication.serial import ThreespaceSerialComClass

#Create a sensor by auto detecting a ThreespaceSerialComClass
sensor = ThreespaceSensor(ThreespaceSerialComClass)

#-------------------------------READING SETTINGS---------------------------

print()
print("Reading settings:")
#If only one value is passed, the result is a singular object. Results are returned as strings
checksum_enabled = sensor.get_settings("header_checksum")
print("Checksum enabled:", bool(checksum_enabled))

#If multiple values are passed, the result is a dictionary with the key as the setting key, and the value as the result
result = sensor.get_settings("debug_mode", "debug_level", "debug_module")
print("Multi-Response:", result)

#With multiple values it is the same but the response is stored still using the key that was requested. This way, you can easily see what key failed.
result = sensor.get_settings("IDontExist")
print("Error Response:", result)

#If a key is invalid in the multiple response format
result = sensor.get_settings("debug_mode", "I Dont Exist", "debug_module", "And neither do I")
print("Multi-Error:", result)


#You can also get bulk keys such as ?settings, ?all... Or querys
#NOTE: If streaming, these bulk setting types will not work via the API unless a normal singular setting is sent first in the list.
#This is because the API requires a known identifier key for the start of the message to parse it out from the streaming data,
#but these have no known static key.
print("?Settings:")
print(sensor.get_settings("settings"))
print("Query ODR")
print(sensor.get_settings("{odr}"))

sensor.cleanup()


from yostlabs.tss3.api import ThreespaceSensor
from yostlabs.communication.serial import ThreespaceSerialComClass

#Create a sensor by auto detecting a ThreespaceSerialComClass
sensor = ThreespaceSensor(ThreespaceSerialComClass)

result = sensor.getPrimaryCorrectedAccelVec()

#Read out the result
accel_vec = result.data

#The result is a dataclass that allows easily accessing 
# individual parts of the data and multiple ways of interpreting it
print("Result:")
#The base ThreespaceCmdResult dataclass
print(result)
#The actual byte response from the sensor that made this result (header + data)
print(result.raw_binary)
print()
print("Header:")
#The ThreespaceHeader data class
print(result.header)
#The individual components of the header as an array
#similar to how the old API functioned
print(result.header.raw)
#The actual byte representation of the header from the sensor
print(result.header.raw_binary)
print()
print("Data:")
print(f"{result.data=}")
print(f"{result.raw_data=}")

sensor.cleanup()


import yostlabs.math.quaternion as quat
import yostlabs.math.vector as vec

import numpy as np
from dataclasses import dataclass
import copy

class ThreespaceGradientDescentCalibration:

    @dataclass
    class StageInfo:
        start_vector: int
        end_vector: int
        stage: int
        scale: float

        count: int = 0

    MAX_SCALE = 1000000000
    MIN_SCALE = 1
    STAGES = [
        StageInfo(0, 6, 0, MAX_SCALE),
        StageInfo(0, 12, 1, MAX_SCALE),
        StageInfo(0, 24, 2, MAX_SCALE)
    ]

    #Note that each entry has a positive and negative vector included in this list
    CHANGE_VECTORS = [
        np.array([0,0,0,0,0,0,0,0,0,.0001,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,0,-.0001,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,0,0,.0001,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,0,0,-.0001,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,0,0,0,.0001], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,0,0,0,-.0001], dtype=np.float64), #First 6 only try to change the bias
        np.array([.001,0,0,0,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([-.001,0,0,0,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,.001,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,-.001,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,.001,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,0,-.001,0,0,0], dtype=np.float64), #Next 6 only try to change the scale
        np.array([0,.0001,0,0,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,-.0001,0,0,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,.0001,0,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,-.0001,0,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,.0001,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,-.0001,0,0,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,.0001,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,-.0001,0,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,.0001,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,-.0001,0,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,.0001,0,0,0,0], dtype=np.float64),
        np.array([0,0,0,0,0,0,0,-.0001,0,0,0,0], dtype=np.float64), #Next 12 only try to change the shear
    ]

    def __init__(self, relative_sensor_orients: list[np.ndarray[float]], no_inverse=False):
        """
        Params
        ------
        relative_sensor_orients : The orientation of the sensor during which each sample is taken if it was tared as if pointing into the screen. 
        The inverse of these will be used to calculate where the axes should be located relative to the sensor
        no_inverse : The relative_sensor_orients will be treated as the sample_rotations
        """
        if no_inverse:
            self.rotation_quats = relative_sensor_orients
        else:
            self.rotation_quats = [np.array(quat.quat_inverse(orient)) for orient in relative_sensor_orients]

    def apply_parameters(self, sample: np.ndarray[float], params: np.ndarray[float]):
        bias = params[9:]
        scale = params[:9]
        scale = scale.reshape((3, 3))
        return scale @ (sample + bias)

    def rate_parameters(self, params: np.ndarray[float], samples: list[np.ndarray[float]], targets: list[np.ndarray[float]]):
        total_error = 0
        for i in range(len(samples)):
            sample = samples[i]
            target = targets[i]

            sample = self.apply_parameters(sample, params)
            
            error = target - sample
            total_error += vec.vec_len(error)
        return total_error

    def generate_target_list(self, origin: np.ndarray):
        targets = []
        for orient in self.rotation_quats:
            new_vec = np.array(quat.quat_rotate_vec(orient, origin), dtype=np.float64)
            targets.append(new_vec)
        return targets

    def __get_stage(self, stage_number: int):
        if stage_number >= len(self.STAGES):
            return None
        #Always get a shallow copy of the stage so can modify without removing the initial values
        return copy.copy(self.STAGES[stage_number])

    def calculate(self, samples: list[np.ndarray[float]], origin: np.ndarray[float], verbose=False, max_cycles_per_stage=1000):
        targets = self.generate_target_list(origin)
        initial_params = np.array([1,0,0,0,1,0,0,0,1,0,0,0], dtype=np.float64)
        stage = self.__get_stage(0)

        best_params = initial_params
        best_rating = self.rate_parameters(best_params, samples, targets)
        count = 0
        while True:
            last_best_rating = best_rating
            params = best_params

            #Apply all the changes to see if any improve the result
            for change_index in range(stage.start_vector, stage.end_vector):
                change_vector = self.CHANGE_VECTORS[change_index]
                new_params = params + (change_vector * stage.scale)
                rating = self.rate_parameters(new_params, samples, targets)

                #A better rating, store it
                if rating < best_rating:
                    best_params = new_params
                    best_rating = rating
            
            if verbose and count % 100 == 0:
                print(f"Round {count}: {best_rating=} {stage=}")
            
            #Decide if need to go to the next stage or not
            count += 1
            stage.count += 1
            if stage.count >= max_cycles_per_stage:
                stage = self.__get_stage(stage.stage + 1)
                if stage is None:
                    if verbose: print("Done from reaching count limit")
                    break
                if verbose: print("Going to next stage from count limit")
                
            if best_rating == last_best_rating: #The rating did not improve
                if stage.scale == self.MIN_SCALE: #Go to the next stage since can't get any better in this stage!
                    stage = self.__get_stage(stage.stage + 1)
                    if stage is None:
                        if verbose: print("Done from exhaustion")
                        break
                    if verbose: print("Going to next stage from exhaustion")
                else:   #Reduce the size of the changes to hopefully get more accurate tuning
                    stage.scale *= 0.1  
                    if stage.scale < self.MIN_SCALE:
                        stage.scale = self.MIN_SCALE
            else: #Rating got better! To help avoid falling in a local minimum, increase the size of the change to see if that could make it better
                stage.scale *= 1.1
        
        if verbose:
            print(f"Final Rating: {best_rating}")
            print(f"Final Params: {best_params}")

        return best_params
    
def fibonacci_sphere(samples=1000):
    points = []
    phi = np.pi * (3. - np.sqrt(5.))  # golden angle

    for i in range(samples):
        y = 1 - (i / float(samples - 1)) * 2  # y goes from 1 to -1
        radius = np.sqrt(1 - y * y)
        theta = phi * i
        x = np.cos(theta) * radius
        z = np.sin(theta) * radius
        points.append((x, y, z))

    return np.array(points)

class ThreespaceSphereCalibration:

    def __init__(self, max_comparison_points=500):
        self.buffer = 0.1
        self.test_points = fibonacci_sphere(samples=max_comparison_points)
        self.clear()

    def process_point(self, raw_mag: list[float]):
        if len(self.samples) == 0:
            self.samples = np.array([raw_mag])
            return True
        
        raw_mag = np.array(raw_mag, dtype=np.float64)
        new_len = np.linalg.norm(raw_mag)

        avg_len = np.linalg.norm(self.samples, axis=1)
        avg_len = (avg_len + new_len) / 2 * self.buffer

        dist = np.linalg.norm(self.samples - raw_mag, axis=1)
        if np.any(dist < avg_len):
            return False
        
        self.samples = np.concatenate((self.samples, [raw_mag]))
        self.__update_density(raw_mag / new_len)
        return True
    
    def __update_density(self, normalized_point: np.ndarray):
        #First check to see if the new point is the closest point for any of the previous points
        dots = np.sum(self.test_points * normalized_point, axis=1)
        self.closest_dot = np.maximum(self.closest_dot, dots)
        self.largest_delta_index = np.argmin(self.closest_dot)
        self.largest_delta = np.rad2deg(np.acos(self.closest_dot[self.largest_delta_index]))

    def clear(self):
        self.samples = np.array([])
        self.closest_dot = np.array([-1] * len(self.test_points))
        self.largest_delta = 180
        self.largest_delta_index = 0

    @property
    def sparsest_vector(self):
        return self.test_points[self.largest_delta_index]

    def calculate(self):
        """
        Returns matrix and bias
        """
        comp_A, comp_b, comp_d = ThreespaceSphereCalibration.alternating_least_squares(np.array(self.samples))
        comp_A, comp_b = ThreespaceSphereCalibration.make_calibration_params(comp_A, comp_b, comp_d, 1)

        comp_scale: float = comp_A[0][0] + comp_A[1][1] + comp_A[2][2]
        comp_scale /= 3

        comp_A /= comp_scale

        return comp_A.flatten().tolist(), comp_b.tolist()

    @staticmethod
    def alternating_least_squares(data: np.ndarray[np.ndarray]) -> tuple[np.ndarray, np.ndarray, float]:
        n = 3
        m = len(data)

        #Step 1
        noise_variance = 0.1
        sigma2 = noise_variance ** 2
        rotini = data.T
        rotini2 = rotini ** 2

        T = np.zeros((5,n,m))
        T[0,:,:] = 1
        T[1,:,:] = rotini
        T[2,:,:] = rotini2 - sigma2
        T[3,:,:] = rotini2 * rotini - 3 * rotini * sigma2
        T[4,:,:] = rotini2 * rotini2 - 6 * rotini2 * sigma2 + 3 * sigma2 * sigma2

        #Step 2
        one = np.ones((n+1,))
        i = np.arange(1, n+1)
        i = np.append(i, 0)

        m1 = np.outer(i.T, one)
        m2 = np.outer(one.T, i)

        M = np.array([ThreespaceSphereCalibration.vec_s(m1), ThreespaceSphereCalibration.vec_s(m2)])

        #Step 3
        nb = int((n + 1) * n / 2 + n + 1)
        R = np.zeros((nb,nb,n), dtype=np.int32)
        
        for p in range(nb):
            for q in range(p, nb):
                for i in range(1, n+1):
                    R[p,q,i-1] = int(M[0,p] == i) + int(M[1,p] == i) + int(M[0,q] == i) + int(M[1,q] == i)

        #Step 4
        nals = np.zeros((nb, nb))
        for p in range(nb):
            for q in range(p, nb):
                sum = 0
                for l in range(m):
                    prod = 1
                    for i in range(n):
                        prod *= T[R[p,q,i],i,l]
                    sum += prod
                nals[p,q] = sum

        #Step 5
        D2 = [i * (i + 1) / 2 for i in range(1, n+1)]
        D = [d for d in range(1, (n+1) * n // 2 + 1) if not d in D2]

        #Step 6
        menorah_als = np.zeros((nb, nb))
        
        for p in range(nb):
            for q in range(p, nb):
                coeff = 2
                if p + 1 in D and q + 1 in D:
                    coeff = 4
                elif not p + 1 in D and not q + 1 in D:
                    coeff = 1
                menorah_als[p,q] = coeff * nals[p,q]
        
        # Fill the lower triangle with the upper triangle values
        i_lower, j_lower = np.tril_indices(nb, k=-1) 
        menorah_als[i_lower, j_lower] = menorah_als[j_lower, i_lower]

        #Step 7
        eigenmat = menorah_als

        #It is unclear if this is correct, there are differences in sign and positions
        eigenvalues, eigenvectors = np.linalg.eig(eigenmat)
        eigenvectors = eigenvectors.T

        #Looks like this section gets the eigen vector with the largest eigen value?
        combinedmatr = [[abs(eigenvalues[i]), eigenvectors[i]] for i in range(len(eigenvalues))]
        combinedmatr.sort(key=lambda a: a[0], reverse=True)

        bals = combinedmatr[-1][1]

        #Step 8 : ensure normalized
        bals = bals / np.linalg.norm(bals)

        #Step 9:
        triangle = n*(n+1)//2
        A = ThreespaceSphereCalibration.inv_vec_s(bals[:triangle])
        b = bals[triangle:nb-1]
        d = bals[-1]

        return A, b, d

    @staticmethod
    def make_calibration_params(Q: np.ndarray, u: np.ndarray, k: float, H_m: float) -> tuple[np.ndarray,np.ndarray]:
        pa = np.linalg.inv(Q)
        pb = u.T
        b = np.dot(pa, pb) * 0.5

        eigenvalues, V = np.linalg.eig(Q)
        D = np.diag(eigenvalues)

        vu_prod = np.dot(V.T, u.T)
        p1a = np.dot(vu_prod.T, np.linalg.inv(D))
        p1b = np.dot(p1a, vu_prod)
        p1 = p1b - (4 * k)

        alpha = 4 * (H_m ** 2) / p1

        aD = np.diag(abs(alpha * eigenvalues) ** 0.5)

        A = np.dot(np.dot(V, aD), V.T)
        
        return A, b

    @staticmethod
    def vec_s(matrix: np.ndarray):
        rows, cols = np.tril_indices(matrix.shape[0])
        return matrix[rows, cols]

    @staticmethod
    def inv_vec_s(vec: np.ndarray):
        #Its unclear if this function works as intended. But this is how the suite does it.
        size = int((-1 + (1 + 8 * len(vec)) ** 0.5) / 2)
        matr = np.zeros((size,size))
        base = 0
        for i in range(size):
            for j in range(i):
                matr[i,j] = vec[base + j]
                matr[j,i] = vec[base + j]
            matr[i,i] = vec[base + i]
            base += i + 1
        
        return matr

    @property
    def num_points(self):
        return len(self.samples)
